import numpy as np
import numpy.linalg as nl
import scipy.spatial.distance as ssd
import scipy.linalg.interpolative as sli

from cs598apk.qbx import QBXExpansion
from cs598apk.utils import cumsum, interp_decomp


class CompressedMatrix:
    def __init__(self, A, **kwargs):
        self.A = A
        self.parent = None

        self.child = kwargs.pop("child", None)
        self.epsilon = kwargs.get("epsilon", 1.0e-5)
        self.rank = kwargs.get("rank", None)
        self.nranges = kwargs.get("nranges", None)
        self.maxlevel = int(kwargs.get("maxlevel", 1))

        if self.child is None:
            self.level = self.maxlevel
        else:
            self.level = self.child.level - 1

        if self.child is None:
            if self.nranges is None:
                self.nranges = 2 ** self.maxlevel
            else:
                assert int(np.log2(self.nranges)) >= self.maxlevel
        else:
            self.nranges = self.child.nranges // 2

        # compute indices for current level from child (clustering)
        self._setup_level_index()
        # allocate compressed matrices
        self._setup_compress()
        # compress matrix
        self.compress_matrix(**kwargs)
        # compute ranges in compressed matrix
        self._setup_ranges()
        # compute diagonal matrix in decomposition
        self._setup_diagonal()

        if kwargs.get("verbose", True):
            self._log()

    def _setup_level_index(self):
        if self.child is None:
            nrows, ncols = self.A.shape
            self.rindex = \
                np.array(np.array_split(np.arange(0, nrows), self.nranges))
            self.cindex = \
                np.array(np.array_split(np.arange(0, ncols), self.nranges))
        else:
            self.rindex = np.empty(self.nranges, dtype=object)
            self.cindex = np.empty(self.nranges, dtype=object)

            for i in range(self.nranges):
                r = np.s_[2 * i:2 * (i + 1)]

                self.rindex[i] = \
                    np.hstack(self.child.compressed_rindex[r]).astype(int)
                self.cindex[i] = \
                    np.hstack(self.child.compressed_cindex[r]).astype(int)

        # construct row and column block indices
        mask = np.arange(0, self.nranges)

        self.block_rindex = np.zeros(self.nranges, dtype=object)
        self.block_cindex = np.zeros(self.nranges, dtype=object)
        for i in range(self.nranges):
            self.block_rindex[i] = \
                np.hstack(self.cindex[mask != i]).astype(int)
            self.block_cindex[i] = \
                np.hstack(self.rindex[mask != i]).astype(int)

    def _setup_compress(self):
        # allocate block matrices for decomposition:
        #   A = L S R + D
        self.L = np.empty((self.nranges, self.nranges), dtype=object)
        self.R = np.empty((self.nranges, self.nranges), dtype=object)
        self.D = np.empty((self.nranges, self.nranges), dtype=object)
        self.M = np.empty((self.nranges, self.nranges), dtype=object)

        # allocate arrays for compressed indices
        self.compressed_rindex = np.zeros(self.nranges, dtype=object)
        self.compressed_cindex = np.zeros(self.nranges, dtype=object)

    def _setup_ranges(self):
        self.rranges = cumsum(self.rindex)
        self.cranges = cumsum(self.cindex)
        self.compressed_rranges = cumsum(self.compressed_rindex)
        self.compressed_cranges = cumsum(self.compressed_cindex)

        self.shape = (self.rranges[-1], self.cranges[-1])
        self.cshape = (self.compressed_rranges[-1],
                       self.compressed_cranges[-1])

    def _setup_diagonal(self):
        if self.level == self.maxlevel:
            for i in range(self.nranges):
                block = np.ix_(self.rindex[i], self.cindex[i])
                self.D[i, i] = nl.inv(self.A[block])
        else:
            block = np.empty((2, 2), dtype=object)
            for i in range(self.nranges):
                block01 = np.ix_(self.child.compressed_rindex[2 * i + 0],
                                 self.child.compressed_cindex[2 * i + 1])
                block10 = np.ix_(self.child.compressed_rindex[2 * i + 1],
                                 self.child.compressed_cindex[2 * i + 0])

                block[0, 0] = self.child.M[2 * i, 2 * i]
                block[0, 1] = self.A[block01]
                block[1, 0] = self.A[block10]
                block[1, 1] = self.child.M[2 * i + 1, 2 * i + 1]

                self.D[i, i] = nl.inv(np.block(block.tolist()))

        for i in range(self.nranges):
            self.M[i, i] = \
                nl.inv(self.R[i, i] @ self.D[i, i] @ self.L[i, i])

    def _log(self):
        print("[{:02}] rank {}".format(self.level, self.rank))
        print("[{:02}] blocks {} x {}"
              .format(self.level, len(self.rindex), len(self.cindex)))
        print("[{:02}] shape {} compressed shape {}"
              .format(self.level, self.shape, self.cshape))

    def solve(self, b):
        def ranges(cr, r):
            for j in range(self.nranges):
                yield np.s_[cr[j]:cr[j + 1]], np.s_[r[j]:r[j + 1]], j

        # compress right-hand side
        bbar = np.empty(self.cshape[1])
        for cr, r, i in ranges(self.compressed_cranges, self.cranges):
            bbar[cr] = self.M[i, i] @ self.R[i, i] @ self.D[i, i] @ b[r]

        # solve system
        if self.parent is None:
            mask = np.arange(0, self.nranges)

            # construct dense matrix
            B = np.empty((self.nranges, self.nranges), dtype=object)
            for i in range(self.nranges):
                B[i, i] = self.M[i, i]

                for j in mask[mask != i]:
                    block = np.ix_(self.compressed_rindex[i],
                                   self.compressed_cindex[j])

                    B[i, j] = self.A[block]

            # direct solve
            B = np.block(B.tolist())
            y = nl.solve(B, bbar)
        else:
            # solve reduced system
            y = self.parent.solve(bbar)

        # back substitute
        x = np.empty(self.shape[0])
        for cr, r, i in ranges(self.compressed_rranges, self.rranges):
            x[r] = b[r] - \
                   self.L[i, i] @ bbar[cr] + \
                   self.L[i, i] @ self.M[i, i] @ y[cr]
            x[r] = self.D[i, i] @ x[r]

        return x


class InterpolativeMatrix(CompressedMatrix):
    def __init__(self, A, **kwargs):
        super().__init__(A, **kwargs)
        if kwargs.get("verbose", True):
            print("Compressed level {}".format(self.level))

        if self.level > 1:
            kwargs["child"] = self
            self.parent = InterpolativeMatrix(A, **kwargs)

    def compress_matrix(self, **kwargs):
        for i in range(self.nranges):
            k = self.rank

            # construct column decomposition
            block = np.ix_(self.block_cindex[i], self.cindex[i])
            k, idx, proj = interp_decomp(self.A[block], k, self.epsilon)

            self.R[i, i] = sli.reconstruct_interp_matrix(idx, proj)
            self.compressed_cindex[i] = self.cindex[i][idx[:k]]

            # construct row decomposition
            block = np.ix_(self.rindex[i], self.block_rindex[i])
            k, idx, proj = interp_decomp(self.A[block].T, k, self.epsilon)

            self.L[i, i] = sli.reconstruct_interp_matrix(idx, proj).T
            self.compressed_rindex[i] = self.rindex[i][idx[:k]]


class SkeletonMatrix(CompressedMatrix):
    def __init__(self, A, **kwargs):
        super().__init__(A, **kwargs)
        if kwargs.get("verbose", True):
            print("Compressed level {}".format(self.level))

        if self.level > 1:
            kwargs["child"] = self
            self.parent = SkeletonMatrix(A, **kwargs)

    def compress_matrix(self, **kwargs):
        self.nproxies = int(kwargs.get("nproxies"))
        self.sources = kwargs.get("sources")
        self.kernel = kwargs.get("kernel")
        self.ratio = kwargs.get("ratio", 1.25)
        self.limit = kwargs.get("limit", QBXExpansion.Boundary)

        # create decomposition:
        for i in range(self.nranges):
            # construct column decomposition
            A = self._skeletonize(i, self.sources, self.kernel, row=False)
            k, idx, proj = interp_decomp(A, self.rank, self.epsilon)

            self.R[i, i] = sli.reconstruct_interp_matrix(idx, proj)
            self.compressed_cindex[i] = self.cindex[i][idx[:k]]

            # construct row decomposition
            A = self._skeletonize(i, self.sources, self.kernel, row=True)
            k, idx, proj = interp_decomp(A.T, k, self.epsilon)

            self.L[i, i] = sli.reconstruct_interp_matrix(idx, proj).T
            self.compressed_rindex[i] = self.rindex[i][idx[:k]]

    def _skeletonize(self, i, source, kernel, row=True):
        block = self.rindex[i] if row else self.cindex[i]
        offblock = self.block_rindex[i] if row else self.block_cindex[i]

        # compute block center of mass
        center = np.mean(source.x[:, block], axis=1).reshape(-1, 1)

        # compute block proxy radius
        ri = re = 0.0
        if self.limit == QBXExpansion.NoExpansion:
            ri = np.max(ssd.cdist(source.x[:, block].T, source.x[:, block].T))
        else:
            # NOTE: need to make sure the proxy points are outside any
            # expansion centers.
            if self.limit & QBXExpansion.Interior:
                ri = np.max(nl.norm(center - source.ci[:, block], axis=0) +
                            source.ratio * source.h[block])
            if self.limit & QBXExpansion.Exterior:
                re = np.max(nl.norm(center - source.ce[:, block], axis=0) +
                            source.ratio * source.h[block])
        r = self.ratio * max(ri, re)

        # construct proxies
        theta = np.linspace(0.0, 2.0 * np.pi, self.nproxies)
        proxies = center + \
            r * np.array([np.cos(theta), np.sin(theta)])

        # construct additional proxies from nearby blocks
        mask = np.full(source.x.shape[1], False)
        mask[offblock] = \
            nl.norm(center - source.x[:, offblock], axis=0) <= r

        # construct interaction matrix
        if row:
            C = source.domain_operator(kernel, proxies, rindex=block)
            C = np.hstack([C, self.A[block, :][:, mask]])
        else:
            C = source.domain_operator(kernel, proxies, cindex=block)
            C = np.vstack([C, self.A[mask, :][:, block]])

        return C
