import numpy as np
import sympy as sp

import itertools


class GreenFunction:
    def __init__(self, ndim=2):
        self.ndim = ndim

        # define target point symbols
        self.x = [sp.Symbol("x{}".format(i)) for i in range(ndim)]
        # define source point symbols
        self.y = [sp.Symbol("y{}".format(i)) for i in range(ndim)]

        self.variables = [*self.x, *self.y]
        self.fn = None

    def source(self):
        return self.y

    def target(self):
        return self.x

    def symbolic(self):
        return None

    def green(self):
        return self

    def __call__(self, **kwargs):
        if self.fn is None:
            self.fn = sp.lambdify(self.variables, self.kernel, "numpy")

        args = []
        for v in self.variables:
            args.append(kwargs.get(v.name))

        return self.fn(*args)


class GreenLaplace(GreenFunction):
    def __init__(self, ndim=2):
        super().__init__(ndim=ndim)

        r = sp.sqrt(sum([(x - y) ** 2 for x, y in zip(self.x, self.y)]))
        if ndim == 2:
            self.kernel = sp.log(r) / (2.0 * sp.pi)
        elif ndim == 3:
            self.kernel = 1.0 / (4.0 * sp.pi * r)
        else:
            raise ValueError("Unsupported dimension: ndim = {}".format(ndim))

    def symbolic(self):
        return self.kernel


class PotentialLayerKernel:
    def __init__(self, K):
        self.K = K
        self.fn = None

        self.ndim = K.ndim
        self.variables = K.variables[::]
        self.kernel = K.symbolic()

    def source(self):
        return self.K.source()

    def target(self):
        return self.K.target()

    def symbolic(self):
        return self.kernel

    def green(self):
        return self.K.green()

    def __call__(self, **kwargs):
        if self.fn is None:
            self.fn = sp.lambdify(self.variables, self.kernel, "numpy")

        args = []
        for v in self.variables:
            args.append(kwargs.get(v.name))

        return self.fn(*args)


class SingleLayerKernel(PotentialLayerKernel):
    def __init__(self, K):
        super().__init__(K)


class DoubleLayerKernel(PotentialLayerKernel):
    def __init__(self, K):
        super().__init__(K)

        # define source points
        y = self.source()
        # define source normal
        ny = [sp.Symbol("ny{}".format(i)) for i in range(K.ndim)]
        self.variables.extend(ny)

        # define kernel
        self.kernel = ny[0] * self.kernel.diff(y[0]) + \
                      ny[1] * self.kernel.diff(y[1])


class ExpandedKernel(PotentialLayerKernel):
    def __init__(self, kernel, order=3, local=True):
        super().__init__(kernel)
        self.order = order
        self.local = local

        # define expansion center
        c = [sp.Symbol("c{}".format(i)) for i in range(kernel.ndim)]
        self.variables.extend(c)

        # compute set of multi-indices
        self.indices = []
        for i, j in itertools.product(range(order + 1), repeat=2):
            if (i + j) <= order:
                self.indices.append(np.array([i, j], dtype=int))
        self.nindices = len(self.indices)

        # compute expansion derivatives
        if local:
            x = kernel.target()
        else:
            x = kernel.source()

        self.derivatives = np.empty(self.nindices, dtype=object)
        for i in range(self.nindices):
            K = kernel.symbolic()
            for mi, v in zip(self.indices[i], x):
                K = K.diff(v, mi)

            self.derivatives[i] = K

        # compute expansion terms
        self.terms = np.empty(self.nindices, dtype=object)
        for i, p in enumerate(self.indices):
            T = 1.0
            for j in range(kernel.ndim):
                T *= (x[j] - c[j]) ** p[j] / sp.factorial(p[j])

            if not local:
                T = T * (-1.0) ** np.prod(p)

            self.terms[i] = \
                T * self.derivatives[i].subs(x[0], c[0]).subs(x[1], c[1])

        self.kernel = sum(self.terms)
