import enum

import numpy as np
import numpy.linalg as nl
import scipy.special as ss


class QBXExpansion(enum.IntFlag):
    NoExpansion = 0
    Interior = 1
    Exterior = 2
    Boundary = 3


class QBXPanel:
    def __init__(self, geometry, a, b, nodes, weights, ratio=0.5):
        self.a = a
        self.b = b
        self.refine = False

        # compute arclength points
        self.t = 0.5 * (a + b) + 0.5 * (b - a) * nodes
        # compute quadrature points
        self.x = geometry.points(self.t)
        # compute normals
        g = geometry.gradient(self.t)
        J = nl.norm(g, 2, axis=0)
        self.n = np.array([g[1, :], -g[0, :]]) / J

        # compute weights
        self.w = 0.5 * weights * J * (b - a)
        # compute panel length
        self.h = np.sum(self.w)

        # compute expansion centers
        self.cint = self.x - ratio * self.h * self.n
        self.cext = self.x + ratio * self.h * self.n
        # compute center of mass
        self.m = np.sum(self.x, axis=1) / self.x.shape[1]
        # compute panel bounding box
        self.bbox = np.array([np.min(self.x, axis=1),
                              np.max(self.x, axis=1)])


class QBX:
    def __init__(self, geometry, npanels, order=7, ratio=0.5, adapt=False):
        self.geometry = geometry
        self.shape = (npanels, order + 1)
        self.ratio = ratio

        self._quadrature()
        if adapt:
            self._adapt()
        self._flatten()

    def _quadrature(self):
        # panel endpoints
        intervals = np.linspace(0.0, 2.0 * np.pi, self.shape[0] + 1)
        # Gauss-Legendre quadrature
        info = ss.legendre(self.shape[1]).weights
        self.nodes = info[:, 0]
        self.weights = info[:, 2]

        # define panels
        self.panels = []
        for i in range(self.shape[0]):
            panel = QBXPanel(self.geometry, *intervals[i:i + 2],
                             self.nodes, self.weights, self.ratio)
            self.panels.append(panel)

    def _flatten(self):
        self._ranges = np.arange(0, np.prod(self.shape) + 1, self.shape[1])

        self.t = np.hstack([p.t for p in self.panels])
        self.x = np.hstack([p.x for p in self.panels])
        self.n = np.hstack([p.n for p in self.panels])
        self.w = np.hstack([p.w for p in self.panels])
        self.h = np.hstack([np.full(self.shape[1], p.h) for p in self.panels])

        self.ci = np.empty(self.x.shape)
        self.ce = np.empty(self.x.shape)
        for r, i in self.iterranges():
            self.ci[:, r] = \
                self.x[:, r] - self.ratio * self.h[r] * self.n[:, r]
            self.ce[:, r] = \
                self.x[:, r] + self.ratio * self.h[r] * self.n[:, r]

        del self.panels

    def _adapt(self):
        m = 0
        while self._mark_panels() and m < 10:
            self._refine_panels()
            m = m + 1

    def _mark_panels(self):
        def centers(p):
            for i in range(self.shape[1]):
                yield p.cint[:, i], p.cext[:, i]

        def distance(c, p):
            return np.min(nl.norm(c.reshape(-1, 1) - p.x, 2, axis=0))

        # mark panels for refinement
        # TODO: should really make use of the quadtree for this.
        refinement_required = False
        for p in self.panels:
            for cint, cext in centers(p):
                for panel in self.panels:
                    if p is panel:
                        continue

                    if distance(cint, panel) <= self.ratio * p.h or \
                       distance(cext, panel) <= self.ratio * p.h:
                        p.refine = True
                        refinement_required = True

                    if p.refine:
                        break
                if p.refine:
                    break

        return refinement_required

    def _refine_panels(self):
        panels = []
        for panel in self.panels:
            if not panel.refine:
                panels.append(panel)
                continue

            intervals = np.linspace(panel.a, panel.b, 5)
            for i in range(4):
                p = QBXPanel(self.geometry, *intervals[i:i + 2],
                             self.nodes, self.weights, self.ratio)
                panels.append(p)
        print("Refinement: {} {}".format(self.shape[0], len(panels)))

        self.panels = panels
        self.shape = (len(panels), self.shape[1])

    def ranges(self, i):
        return range(self._ranges[i], self._ranges[i + 1])

    def iterranges(self):
        for i in range(self.shape[0]):
            yield self.ranges(i), i

    def boundary_operator(self, K, rindex=None, cindex=None,
                          limit=QBXExpansion.Interior):
        s = 0.5 if limit == QBXExpansion.Boundary else 1.0
        if rindex is None:
            rindex = np.s_[:]
        if cindex is None:
            cindex = np.s_[:]

        # setup kernel arguments
        args = {
            "x0": self.x[0, rindex].reshape(-1, 1),
            "x1": self.x[1, rindex].reshape(-1, 1),
            "y0": self.x[0, cindex].reshape(1, -1),
            "y1": self.x[1, cindex].reshape(1, -1),
            "ny0": self.n[0, cindex].reshape(1, -1),
            "ny1": self.n[1, cindex].reshape(1, -1)
            }

        L = 0.0
        if limit & QBXExpansion.Interior:
            args["c0"] = self.ci[0, rindex].reshape(-1, 1)
            args["c1"] = self.ci[1, rindex].reshape(-1, 1)

            L += s * K(**args)
        if limit & QBXExpansion.Exterior:
            args["c0"] = self.ce[0, rindex].reshape(-1, 1)
            args["c1"] = self.ce[1, rindex].reshape(-1, 1)

            L += s * K(**args)

        return L * self.w[cindex]

    def domain_operator(self, K, x, rindex=None, cindex=None):
        # setup kernel arguments
        if rindex is not None:
            args = {
                "x0": self.x[0, rindex].reshape(-1, 1),
                "x1": self.x[1, rindex].reshape(-1, 1),
                "y0": x[0, :].reshape(1, -1),
                "y1": x[1, :].reshape(1, -1)
            }

            return K.green()(**args)
        elif cindex is not None:
            args = {
                "x0": x[0, :].reshape(-1, 1),
                "x1": x[1, :].reshape(-1, 1),
                "y0": self.x[0, cindex].reshape(1, -1),
                "y1": self.x[1, cindex].reshape(1, -1),
                "ny0": self.n[0, cindex].reshape(1, -1),
                "ny1": self.n[1, cindex].reshape(1, -1)
            }

            return K.K(**args) * self.w[cindex]
        else:
            raise ValueError("rindex or cindex required")
