import numpy as np
import numpy.linalg as nl
import matplotlib.pyplot as mp
import matplotlib.patches as mpatch

import collections
from cs598apk import utils


class Quadtree:
    def __init__(self, bbox, parent=None, pid=-1, maxelements=10):
        self.bbox = bbox
        self.maxelements = maxelements
        self.parent = parent
        self.pid = pid

        if self.parent is None:
            self.level = 0
        else:
            self.level = self.parent.level + 1

        self.elements = []
        self.children = None

    def insert(self, element, bbox):
        if self.children is None:
            self.elements.append((element, bbox))
            if len(self.elements) > self.maxelements:
                self._subdivide()
        else:
            child = self._find_child(bbox)
            child.insert(element, bbox)

    def _find_child(self, bbox):
        if self.children is None:
            return self

        for child in self.children:
            if child.intersects(bbox):
                return child

        raise ValueError("Could not find child for {} {} in parent {} {}"
                         .format(bbox[:, 0], bbox[:, 1],
                                 self.bbox[:, 0], self.bbox[:, 1]))

    def _subdivide(self):
        # create new children in z order:
        #   *-----*-----*
        #   |  2  |  3  |
        #   |     |     |
        #   *-----*-----*
        #   |     |     |
        #   |  0  |  1  |
        #   *-----*-----*
        radius = 0.5 * self.size()

        self.children = np.empty(2 ** self.bbox.shape[0], dtype=object)
        for i in range(self.children.shape[0]):
            offset = utils.bits(i, 2)
            bbox = np.array([self.bbox[:, 0] + (offset + 0) * radius,
                             self.bbox[:, 0] + (offset + 1) * radius])

            self.children[i] = Quadtree(bbox.reshape(2, 2).T,
                                        parent=self, pid=i,
                                        maxelements=self.maxelements)

        # insert all elements into children
        for e, ebox in self.elements:
            child = self._find_child(ebox)
            child.insert(e, ebox)

        del self.elements

    def build_interactions(self, root=None):
        if root is None:
            root = self

        self.adjacent_boxes = []
        self.separated_boxes = []
        self.smaller_boxes = []
        self.larger_boxes = []

        # build interaction lists for current node
        if self.level > 1:
            size = self.parent.size()
            bbox = np.array([self.parent.bbox[:, 0] - size,
                             self.parent.bbox[:, 1] + size]).reshape(2, 2).T

            self._find_boxes(root, bbox)

        # recurse and build interactions lists for each child
        if not self.is_leaf():
            for child in self.children:
                child.build_interactions(root)

    def _find_boxes(self, box, bbox):
        if not box.intersects(bbox, epsilon=-1.0e-12):
            return

        # check List 2
        if box.level == self.level:
            # compute box radius
            r = nl.norm(0.5 * self.size(), float("inf"))
            # compute center distance
            R = nl.norm(self.center() - box.center(), 2)

            if R > 3.0 * r:
                self.separated_boxes.append(box)
                return

        if box.is_leaf():
            if self is box or not self.is_leaf():
                return

            if box.intersects(self.bbox):
                self.adjacent_boxes.append(box)
            elif box.level < self.level:
                self.larger_boxes.append(box)
            elif box.level > self.level:
                self.smaller_boxes.append(box)
        else:
            for child in box.children:
                self._find_boxes(child, bbox)

    def intersects(self, box, epsilon=1.0e-12):
        if isinstance(box, Quadtree):
            bbox = box.bbox
        else:
            bbox = box

        for i in range(self.bbox.shape[0]):
            d = max(self.bbox[i, 0], bbox[i, 0]) - \
                min(self.bbox[i, 1], bbox[i, 1])

            if d > epsilon:
                return False

        return True

    def is_leaf(self):
        return self.children is None

    def size(self):
        return self.bbox[:, 1] - self.bbox[:, 0]

    def center(self):
        return 0.5 * (self.bbox[:, 0] + self.bbox[:, 1])

    def list_boxes(self, all_boxes=True):
        if self.is_leaf():
            return [self]

        boxes = []
        if all_boxes:
            boxes.append(self)

        for child in self.children:
            boxes.extend(child.list_boxes(all_boxes))

        return boxes

    def list_levels(self, level=0):
        levels = collections.defaultdict(list)
        levels[level].append(self)

        if self.children is None:
            return levels

        for child in self.children:
            child_levels = child.list_levels(level + 1)

            for child_level, boxes in child_levels.items():
                levels[child_level].extend(boxes)

        return levels

    def plot(self, axis=None, **kwargs):
        if axis is None:
            axis = mp.gca()

        self._plot_box(axis, -1, **kwargs)
        if self.children is None:
            return

        for child in self.children:
            child.plot(axis=axis, **kwargs)

    def plot_interactions(self, axis=None, **kwargs):
        if axis is None:
            axis = mp.gca()

        # make sure the rectangles are filled
        kwargs["fill"] = True

        # plot box
        kwargs["color"] = "k"
        self._plot_box(axis, 0, **kwargs)
        # List 1
        kwargs["color"] = "tab:blue"
        for box in self.adjacent_boxes:
            box._plot_box(axis, 1, **kwargs)
        # List 2
        kwargs["color"] = "tab:red"
        for box in self.separated_boxes:
            box._plot_box(axis, 2, **kwargs)
        # List 3
        kwargs["color"] = "tab:green"
        for box in self.smaller_boxes:
            box._plot_box(axis, 3, **kwargs)
        # List 4
        kwargs["color"] = "tab:orange"
        for box in self.larger_boxes:
            box._plot_box(axis, 4, **kwargs)

    def _plot_box(self, axis, list_id, **kwargs):
        x = self.bbox[:, 0]
        w, h = self.size()

        patch = mpatch.Rectangle(x, w, h, **kwargs)
        axis.add_patch(patch)

        if list_id > 0:
            axis.text(*self.center(),
                      "{}".format(list_id), weight="extra bold")

    def __repr__(self):
        return "Box{}".format(self.bbox.reshape(-1))
