import numpy as np
import scipy.linalg as sl

import time
import cs598apk as cs


def curve(t):
    x = (3.0 / 4.0) * np.cos(t - np.pi / 4.0) * (1.0 + np.sin(2.0 * t) / 2.0)
    y = np.sin(t - np.pi / 4.0) * (1.0 + np.sin(4.0 * t) / 2.0)

    return np.array([x, y])


def dcurve_dt(t):
    dx = (3.0 / 4.0) * np.cos(t - np.pi / 4.0) * np.cos(2.0 * t) - \
        (3.0 / 4.0) * np.sin(t - np.pi / 4.0) * (1.0 + np.sin(2.0 * t) / 2.0)
    dy = np.cos(t - np.pi / 4.0) * (1.0 + np.sin(4.0 * t) / 2.0) + \
        np.sin(t - np.pi / 4.0) * np.cos(4.0 * t) * 2.0

    return np.array([dx, dy])


def bench_panel_level(**kwargs):
    # define geometry
    geometry = cs.ParametricGeometry(curve, dcurve_dt)
    # define solution representation
    K = cs.DoubleLayerKernel(cs.GreenLaplace())
    K = cs.ExpandedKernel(K, order=kwargs["p"])

    # define test parameters
    npanels = 2 ** np.arange(7, 12)
    maxlevel = 2 ** np.arange(3)

    solve_np = {}
    decomp_np = {}
    solve_id = {}
    decomp_id = {}
    solve_sk = {}
    decomp_sk = {}

    for n in npanels:
        kwargs["npanels"] = n

        # define quadrature
        qbx = cs.QBX(geometry,
                     kwargs["npanels"],
                     order=kwargs["q"],
                     adapt=kwargs["adapt"])
        kwargs["sources"] = qbx
        kwargs["kernel"] = K

        # define operator
        D = qbx.boundary_operator(K, limit=cs.QBXExpansion.Boundary)
        D = 0.5 * np.eye(D.shape[0]) + D
        # define right-hand side
        b = np.random.random(D.shape[0])

        # solve using numpy
        t_start = time.time()
        LU, P = sl.lu_factor(D)
        t_end = time.time()
        decomp_np[(0, n)] = t_end - t_start
        print("np.decomp:   {:.6f}".format(t_end - t_start))

        t_start = time.time()
        sl.lu_solve((LU, P), b)
        t_end = time.time()
        solve_np[(0, n)] = t_end - t_start
        print("np.solve:    {:.6f}".format(t_end - t_start))

        for l in maxlevel:
            kwargs["maxlevel"] = l
            print(">> {} {}".format(l, n))

            # solve using full ID
            t_start = time.time()
            matrix = cs.InterpolativeMatrix(D, **kwargs)
            t_end = time.time()
            decomp_id[(l, n)] = t_end - t_start
            print("id.decomp:   {:.6f}".format(t_end - t_start))

            t_start = time.time()
            matrix.solve(b)
            t_end = time.time()
            solve_id[(l, n)] = t_end - t_start
            print("id.solve:    {:.6f}".format(t_end - t_start))

            # solve using skeleton ID
            t_start = time.time()
            matrix = cs.SkeletonMatrix(D, **kwargs)
            t_end = time.time()
            decomp_sk[(l, n)] = t_end - t_start
            print("sk.decomp:   {:.6f}".format(t_end - t_start))

            t_start = time.time()
            matrix.solve(b)
            t_end = time.time()
            solve_sk[(l, n)] = t_end - t_start
            print("sk.solve:    {:.6f}".format(t_end - t_start))


def bench_qbx_level(**kwargs):
    pass


if __name__ == "__main__":
    args = {}

    # order of QBX expansion
    args["p"] = 3
    # nodes in Gauss quadrature
    args["q"] = 7
    # number of panels
    args["npanels"] = 256
    # adapt panel geometry for QBX
    args["adapt"] = False
    # maximum level of recursion
    maxlevel = int(np.log2(args["q"] + 1))
    args["maxlevel"] = 3
    # number of ranges
    args["nranges"] = 32
    # interpolative decomposition tolerance
    args["epsilon"] = 1.0e-10
    # number of proxies
    args["nproxies"] = 50
    # size of proxy circle
    args["ratio"] = 1.25
    # turn of verbosity for different solvers
    args["verbose"] = False

    bench_panel_level(**args)
    bench_qbx_level(**args)
