import numpy as np
import numpy.linalg as nl

import cs598apk as cs


def curve(t):
    x = (3.0 / 4.0) * np.cos(t - np.pi / 4.0) * (1.0 + np.sin(2.0 * t) / 2.0)
    y = np.sin(t - np.pi / 4.0) * (1.0 + np.sin(4.0 * t) / 2.0)

    return np.array([x, y])


def dcurve_dt(t):
    dx = (3.0 / 4.0) * np.cos(t - np.pi / 4.0) * np.cos(2.0 * t) - \
        (3.0 / 4.0) * np.sin(t - np.pi / 4.0) * (1.0 + np.sin(2.0 * t) / 2.0)
    dy = np.cos(t - np.pi / 4.0) * (1.0 + np.sin(4.0 * t) / 2.0) + \
        np.sin(t - np.pi / 4.0) * np.cos(4.0 * t) * 2.0

    return np.array([dx, dy])


def test_decomposition(matrix, qbx, threshold):
    node = matrix
    while True:
        mask = np.arange(node.nranges)

        for i in range(node.nranges):
            for j in mask[mask != i]:
                block = np.ix_(node.rindex[i],
                               node.cindex[j])
                A = D[block]
                block = np.ix_(node.compressed_rindex[i],
                               node.compressed_cindex[j])
                S = D[block]

                error = nl.norm(A - node.L[i, i] @ S @ node.R[j, j])
                assert error < threshold, "{} {}".format(error, threshold)

        node = node.parent
        if node is None:
            break


def test_diagonal_cluster(matrix, qbx, threshold):
    node = matrix
    while True:
        mask = np.arange(node.nranges)

        # form compressed matrix
        S = np.empty(node.cshape)
        for i in range(node.nranges):
            rindex = np.arange(node.compressed_rranges[i],
                               node.compressed_rranges[i + 1])
            cindex = np.arange(node.compressed_cranges[i],
                               node.compressed_cranges[i + 1])
            S[np.ix_(rindex, cindex)] = node.M[i, i]

            for j in mask[mask != i]:
                rindex = np.arange(node.compressed_rranges[i],
                                   node.compressed_rranges[i + 1])
                cindex = np.arange(node.compressed_cranges[j],
                                   node.compressed_cranges[j + 1])
                block = np.ix_(node.compressed_rindex[i],
                               node.compressed_cindex[j])

                S[np.ix_(rindex, cindex)] = D[block]

        # check diagonal (clustered)
        for i in range(node.nranges // 2):
            rindex = np.arange(node.parent.rranges[i],
                               node.parent.rranges[i + 1])
            cindex = np.arange(node.parent.cranges[i],
                               node.parent.cranges[i + 1])
            block = np.ix_(rindex, cindex)

            error = nl.norm(np.eye(len(rindex)) -
                            node.parent.D[i, i] @ S[block])
            assert error < threshold, "{}".format(error)

        node = node.parent
        if node.parent is None:
            break


if __name__ == "__main__":
    # order of QBX expansion
    p = 3
    # nodes in Gauss quadrature
    q = 7
    # number of panels
    npanels = 30
    # number of proxy points
    nproxies = 64
    # radius ratio
    ratio = 1.25
    # maximum level
    maxlevel = int(np.log2(q + 1))

    # define geometry
    geometry = cs.ParametricGeometry(curve, dcurve_dt)
    # define solution representation
    K = cs.DoubleLayerKernel(cs.GreenLaplace())
    K = cs.ExpandedKernel(K, order=p)
    # define quadrature
    qbx = cs.QBX(geometry, npanels, order=q, adapt=True)
    # define integral operator
    D = qbx.boundary_operator(K, limit=cs.QBXExpansion.Boundary)
    D = 0.5 * np.eye(D.shape[0]) + D

    matrix = cs.InterpolativeMatrix(D, maxlevel=maxlevel)
    test_decomposition(matrix, qbx, 1.0e-5)
    test_diagonal_cluster(matrix, qbx, 1.0e-12)

    matrix = cs.SkeletonMatrix(D, maxlevel=maxlevel,
                               ratio=ratio, nproxies=nproxies,
                               sources=qbx, kernel=K)
    test_decomposition(matrix, qbx, 1.0e-6)
    test_diagonal_cluster(matrix, qbx, 1.0e-12)
