import numpy as np
import numpy.linalg as nl

import cs598apk as cs


def curve(t):
    x = (3.0 / 4.0) * np.cos(t - np.pi / 4.0) * (1.0 + np.sin(2.0 * t) / 2.0)
    y = np.sin(t - np.pi / 4.0) * (1.0 + np.sin(4.0 * t) / 2.0)

    return np.array([x, y])


def dcurve_dt(t):
    dx = (3.0 / 4.0) * np.cos(t - np.pi / 4.0) * np.cos(2.0 * t) - \
        (3.0 / 4.0) * np.sin(t - np.pi / 4.0) * (1.0 + np.sin(2.0 * t) / 2.0)
    dy = np.cos(t - np.pi / 4.0) * (1.0 + np.sin(4.0 * t) / 2.0) + \
        np.sin(t - np.pi / 4.0) * np.cos(4.0 * t) * 2.0

    return np.array([dx, dy])


def test_solve(matrix, b, xe):
    # solve using numpy
    x0 = np.linalg.solve(matrix.A, b)
    error = nl.norm(xe - x0) / nl.norm(xe)
    residual = nl.norm(matrix.A @ x0 - b) / nl.norm(b)
    print("Error: {:.5e} {:.5e}".format(error, residual))

    # solve using compressed matrix
    x1 = matrix.solve(b)
    error = nl.norm(xe - x1) / nl.norm(xe)
    residual = nl.norm(matrix.A @ x1 - b) / nl.norm(b)
    print("Error: {:.5e} {:.5e}".format(error, residual))


if __name__ == "__main__":
    import time

    # order of QBX expansion
    p = 7
    # nodes in Gauss quadrature
    q = 7
    # number of panels
    npanels = 128
    # maximum level of recursion
    maxlevel = int(np.log2(q + 1))
    maxlevel = 1
    # number of ranges
    nranges = 16
    # interpolative decomposition tolerance
    epsilon = 1.0e-10
    # number of proxies
    nproxies = 30
    # size of proxy circle
    ratio = 1.25

    # define geometry
    geometry = cs.ParametricGeometry(curve, dcurve_dt)
    # define solution representation
    t_start = time.time()
    K = cs.DoubleLayerKernel(cs.GreenLaplace())
    K = cs.GreenLaplace()
    K = cs.ExpandedKernel(K, order=p)
    t_end = time.time()
    print("Time.expansion:  {}".format(t_end - t_start))
    # define quadrature
    t_start = time.time()
    qbx = cs.QBX(geometry, npanels, order=q, adapt=False)
    t_end = time.time()
    print("Time.qbx:        {}".format(t_end - t_start))
    # define integral operator
    t_start = time.time()
    D = qbx.boundary_operator(K, limit=cs.QBXExpansion.Interior)
    # D = 0.5 * np.eye(D.shape[0]) + D
    t_end = time.time()
    print("Time.operator:   {}".format(t_end - t_start))
    print("Condition number: {}".format(nl.cond(D)))

    matrix = cs.InterpolativeMatrix(D,
                                    maxlevel=maxlevel,
                                    nranges=nranges,
                                    epsilon=epsilon)
    sigma = np.random.random(D.shape[0])
    test_solve(matrix, D @ sigma, sigma)
    sigma = np.sin(qbx.t)
    test_solve(matrix, D @ sigma, sigma)

    matrix = cs.SkeletonMatrix(D,
                               maxlevel=maxlevel,
                               nranges=nranges,
                               epsilon=epsilon,
                               ratio=ratio,
                               nproxies=nproxies,
                               sources=qbx, kernel=K)
    sigma = np.random.random(D.shape[0])
    test_solve(matrix, D @ sigma, sigma)
    sigma = np.sin(qbx.t)
    test_solve(matrix, D @ sigma, sigma)
